@@ -1,46 +1,66 @@
 pub mod codegen;
+pub mod simplify;
 
 use crate::range_map::{Range, RangeMap};
 
 use std::convert::TryFrom;
 
-use fxhash::FxHashMap;
+use fxhash::{FxHashMap, FxHashSet};
 
 /// Deterministic finite automate, parameterized on values of accepting states.
 #[derive(Debug)]
-pub struct DFA<A> {
+pub struct DFA<T, A> {
     // Indexed by `StateIdx`
-    states: Vec<State<A>>,
+    states: Vec<State<T, A>>,
 }
 
 #[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
 pub struct StateIdx(usize);
 
-#[derive(Debug)]
-struct State<A> {
-    // Is this the initial state of a rule set? This is important as failure transitions in initial
-    // states consumes the current character, but failure transitions in other states don't.
-    initial: bool,
-    char_transitions: FxHashMap<char, StateIdx>,
-    range_transitions: RangeMap<StateIdx>,
-    fail_transition: Option<StateIdx>,
-    accepting: Option<A>,
+impl StateIdx {
+    fn map<F>(&self, f: F) -> StateIdx
+    where
+        F: Fn(usize) -> usize,
+    {
+        StateIdx(f(self.0))
+    }
 }
 
-impl<A> State<A> {
-    fn new() -> State<A> {
+#[derive(Debug)]
+struct State<T, A> {
+    // Is this the initial state of a rule set? This is important as failure transitions in initial
+    // states consume the current character, but failure transitions in other states don't.
+    initial: bool,
+    char_transitions: FxHashMap<char, T>,
+    range_transitions: RangeMap<T>,
+    fail_transition: Option<T>,
+    accepting: Option<A>,
+    // Predecessors of the state, used to inline code for a state with one predecessor in the
+    // predecessor's code
+    predecessors: FxHashSet<StateIdx>,
+}
+
+impl<T, A> State<T, A> {
+    fn new() -> State<T, A> {
         State {
             initial: false,
             char_transitions: Default::default(),
             range_transitions: Default::default(),
             fail_transition: None,
             accepting: None,
+            predecessors: Default::default(),
         }
     }
+
+    fn has_no_transitions(&self) -> bool {
+        self.char_transitions.is_empty()
+            && self.range_transitions.is_empty()
+            && self.fail_transition.is_none()
+    }
 }
 
-impl<A> DFA<A> {
-    pub fn new() -> (DFA<A>, StateIdx) {
+impl<A> DFA<StateIdx, A> {
+    pub fn new() -> (DFA<StateIdx, A>, StateIdx) {
         let mut initial_state = State::new();
         initial_state.initial = true;
         (
@@ -79,6 +99,8 @@ impl<A> DFA<A> {
             old,
             next
         );
+
+        self.states[next.0].predecessors.insert(state);
     }
 
     pub fn add_range_transition(
@@ -91,21 +113,37 @@ impl<A> DFA<A> {
         self.states[state.0]
             .range_transitions
             .insert(range_begin, range_end, next);
+
+        self.states[next.0].predecessors.insert(state);
     }
 
     pub fn add_fail_transition(&mut self, state: StateIdx, next: StateIdx) {
         assert!(self.states[state.0].fail_transition.is_none());
         self.states[state.0].fail_transition = Some(next);
+        self.states[next.0].predecessors.insert(state);
     }
 }
 
-impl<A> DFA<A> {
+impl<T, A> DFA<T, A> {
+    fn from_states(states: Vec<State<T, A>>) -> DFA<T, A> {
+        DFA { states }
+    }
+
+    fn into_state_indices(self) -> impl Iterator<Item = (StateIdx, State<T, A>)> {
+        self.states
+            .into_iter()
+            .enumerate()
+            .map(|(state_idx, state)| (StateIdx(state_idx), state))
+    }
+}
+
+impl<A> DFA<StateIdx, A> {
     /// Extend the current DFA with another DFA. The extended DFA's states will be renumbered. This
     /// does not add any transitions from the original DFA states to the extension. Accepting
     /// states of the extension is preserved.
     ///
     /// Returns initial state for the extension in the new DFA.
-    pub fn add_dfa(&mut self, other: DFA<A>) -> StateIdx {
+    pub fn add_dfa(&mut self, other: DFA<StateIdx, A>) -> StateIdx {
         let n_current_states = self.states.len();
 
         for State {
@@ -114,6 +152,7 @@ impl<A> DFA<A> {
             range_transitions,
             fail_transition,
             accepting,
+            predecessors,
         } in other.states
         {
             let mut new_char_transitions: FxHashMap<char, StateIdx> = Default::default();
@@ -138,12 +177,18 @@ impl<A> DFA<A> {
                 new_fail_transition = Some(StateIdx(next.0 + n_current_states));
             }
 
+            let predecessors = predecessors
+                .into_iter()
+                .map(|pred| StateIdx(pred.0 + n_current_states))
+                .collect();
+
             self.states.push(State {
                 initial,
                 char_transitions: new_char_transitions,
                 range_transitions: new_range_transitions,
                 fail_transition: new_fail_transition,
                 accepting,
+                predecessors,
             });
         }
 
@@ -151,7 +196,7 @@ impl<A> DFA<A> {
     }
 }
 
-impl<A> DFA<A> {
+impl<A> DFA<StateIdx, A> {
     #[cfg(test)]
     pub fn simulate(&self, chars: &mut dyn Iterator<Item = char>) -> Option<&A> {
         let mut state = StateIdx(0);
@@ -198,7 +243,7 @@ impl Display for StateIdx {
     }
 }
 
-impl<A> Display for DFA<A> {
+impl<A> Display for DFA<StateIdx, A> {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
         for (state_idx, state) in self.states.iter().enumerate() {
             let State {
@@ -207,6 +252,7 @@ impl<A> Display for DFA<A> {
                 range_transitions,
                 fail_transition,
                 accepting,
+                predecessors: _,
             } = state;
 
             if accepting.is_some() {
